/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define	FE_HD_AUTOROTATE	FALSE	// convert for z-up to y-up

#include "houdini.pmh"

namespace fe
{

void HoudiniDraw::drawPoints(const SpatialVector *vertex,
	const SpatialVector *normal,U32 vertices,
	BWORD multicolor,const Color *color)
{
	if(!gdp())
	{
		feLog("HoudiniDraw::drawPoints gdp is NULL\n");
		return;
	}

	for(U32 index=0;index<vertices;index++)
	{
		const SpatialVector& vert=vertex[index];

#if FE_HD_AUTOROTATE
		UT_Vector3 utPoint3(vert[1],vert[2],vert[0]);
#else
		UT_Vector3 utPoint3(vert[0],vert[1],vert[2]);
#endif

		GEO_Point* pGeoPoint=gdp()->appendPoint();

		pGeoPoint->setPos(utPoint3);
	}
}

void HoudiniDraw::drawLines(const SpatialVector *vertex,
	const SpatialVector *normal,U32 vertices,
	StripMode strip,BWORD multicolor,
	const Color *color)
{
	drawPolygons(vertex,normal,vertices,strip,multicolor,color);
}

void HoudiniDraw::drawTriangles(const SpatialVector *vertex,
	const SpatialVector *normal,const Vector2f *texture,U32 vertices,
	StripMode strip,BWORD multicolor,const Color* color)
{
	//* TODO other strip modes

	if(strip==DrawI::e_discrete)
	{
		for(U32 m=0;m<vertices;m+=3)
		{
			if(color && multicolor)
			{
				drawPolygons(&vertex[m],&normal[m],3,DrawI::e_strip,
						multicolor,&color[m]);
			}
			else
			{
				drawPolygons(&vertex[m],&normal[m],3,DrawI::e_strip,
						multicolor,color);
			}
		}
	}
}

void HoudiniDraw::drawPolygons(const SpatialVector *vertex,
	const SpatialVector *normal,U32 vertices,
	StripMode strip,BWORD multicolor,const Color *color)
{
	if(!gdp())
	{
		feLog("HoudiniDraw::drawPolygons gdp is NULL\n");
		return;
	}

	sp<DrawMode> spDrawMode=drawMode();
	String group;
	if(spDrawMode.isValid())
	{
		group=spDrawMode->group();
	}

	const U32 subCount=(strip==DrawI::e_strip)? vertices: 2;
	for(U32 index=0;index<vertices;index+=subCount)
	{
#ifdef FE_HOUDINI_USE_GA
		GA_PointGroup* pPointGroup=NULL;
#else
		GB_PointGroup* pPointGroup=NULL;
#endif
		if(!group.empty())
		{
			pPointGroup=m_pGdp->findPointGroup(group.c_str());
			if(!pPointGroup)
			{
				pPointGroup=m_pGdp->newPointGroup(group.c_str());
			}
			FEASSERT(pPointGroup);
		}

		int npts=subCount;
		int open=GU_POLY_OPEN;
		int appendPts=1;
		GU_PrimPoly* pGuPrimPoly=
				GU_PrimPoly::build(gdp(),npts,open,appendPts);

		GEO_AttributeOwner owner=GEO_POINT_DICT;
		GEO_AttributeHandle handle;
		if(color)
		{
			handle=m_pGdp->getAttribute(owner,GEO_STD_ATTRIB_DIFFUSE);
			if(!handle.isAttributeValid())
			{
				const float defaultColor[3]={0.0, 0.0, 0.0};
#ifdef FE_HOUDINI_USE_GA
				GA_RWAttributeRef attrCd=
						m_pGdp->addFloatTuple(GA_ATTRIB_POINT, "Cd", 3);
				if(attrCd.isValid())
				{
					attrCd.getAttribute()->setTypeInfo(GA_TYPE_COLOR);
				}
#else
				m_pGdp->addPointAttrib(GEO_STD_ATTRIB_DIFFUSE,
						3*sizeof(float),GB_ATTRIB_FLOAT,&defaultColor);
#endif
				handle=m_pGdp->getAttribute(owner,GEO_STD_ATTRIB_DIFFUSE);
			}
		}

		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const SpatialVector& vert=vertex[index+subIndex];

#if FE_HD_AUTOROTATE
			UT_Vector3 utPoint3(vert[1],vert[2],vert[0]);
#else
			UT_Vector3 utPoint3(vert[0],vert[1],vert[2]);
#endif

			GEO_Point* pGeoPoint=pGuPrimPoly->getVertex(subIndex).getPt();

			pGeoPoint->setPos(utPoint3);

			if(pPointGroup)
			{
				pPointGroup->add(pGeoPoint);
			}

			if(color)
			{
				const SpatialVector& color0=
						color[multicolor? index+subIndex: 0];

				handle.setElement(pGeoPoint);
				UT_Vector4 utVector4(color0.temp());
				handle.setV4(utVector4);
			}

			pGuPrimPoly->setVertex(subIndex, pGeoPoint);
		}
	}
}

} // namespace
