/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "vegetation/vegetation.pmh"

#define FE_TOP_SIGNAL_DEBUG		FALSE

using namespace fe;

TreeOp::TreeOp(void)
{
}

TreeOp::~TreeOp(void)
{
#if FE_TOP_SIGNAL_DEBUG
	Peeker peeker;
	if(m_spSignalerI.isValid())
	{
		m_spSignalerI->peek(peeker);
	}

	const String peek=peeker.str();
	if(!peek.empty())
	{
		feLog("\n~TreeOp\n%s\n",peek.c_str());
	}
#endif
}

void TreeOp::initialize(void)
{
	catalog<String>("icon")="FE_beta";

	catalog<String>("TreeName")="Aspen_Quaking";
	catalog<String>("TreeName","label")="Tree Name";
	catalog<String>("TreeName","hint")="Tree growth configuration"
			" (ignored when input geometry is provided).";

	catalog<SpatialVector>("TreeLocation");
	catalog<String>("TreeLocation","label")="Tree Location";
	catalog<String>("TreeLocation","hint")=
			"Displacement of trunk base from origin.";

	catalog<SpatialVector>("WindDirection");
	catalog<String>("WindDirection","label")="Wind Direction";

	catalog<bool>("YUp")=true;
	catalog<String>("YUp","label")="Y Is Up";
	catalog<String>("YUp","hint")=
			"Indicate that the Y axis points up instead of the Z axis.";

	catalog<SpatialVector>("ColliderCenter")=SpatialVector(0.0,0.0,0.0);
	catalog<String>("ColliderCenter","label")="Collider Center";
	catalog<String>("ColliderCenter","hint")="Center of collider sphere.";

	catalog<Real>("ColliderRadius")=1.0;
	catalog<String>("ColliderRadius","label")="Collider Radius";
	catalog<Real>("ColliderRadius","high")=10.0;
	catalog<Real>("ColliderRadius","max")=1e3;
	catalog<String>("ColliderRadius","hint")="Radius of collider sphere.";

	catalog< sp<Component> >("Input Surface");
	catalog<bool>("Input Surface","optional")=true;

//	catalog< sp<Component> >("Output Surface");
//	catalog<String>("Output Surface","IO")="output";

	m_spSignalerI=registry()->create("SignalerI");
	sp<HandlerI> spSimPlant=registry()->create("*.SimPlant");
	sp<HandlerI> spDrawPlant=registry()->create("*.DrawPlant");
	if(!m_spSignalerI.isValid() || !spSimPlant.isValid() ||
			!spDrawPlant.isValid())
	{
		feX("TreeOp::initialize", "can't create components");
	}
	m_modifierArray.push_back(spSimPlant);
	m_modifierArray.push_back(spDrawPlant);

	m_spScope=registry()->master()->catalog()->catalogComponent(
			"Scope","PlantScope");
	FEASSERT(m_spScope.isValid());

	m_forestRV.bind(m_spScope);
	m_forestRV.createRecord();
	set(m_forestRV.effectorVelocity());
	set(m_forestRV.windSource());

	m_plantRV.bind(m_spScope);

	m_spSignalerI->insert(spSimPlant,m_forestRV.layout());
	m_spSignalerI->insert(spDrawPlant,m_forestRV.layout());
}

SpatialVector TreeOp::convertY(SpatialVector a_vector)
{
	return catalog<bool>("YUp")?
			SpatialVector(a_vector[0],-a_vector[2],a_vector[1]): a_vector;
}

SpatialVector TreeOp::unconvertY(SpatialVector a_vector)
{
	return catalog<bool>("YUp")?
			SpatialVector(a_vector[0],a_vector[2],-a_vector[1]): a_vector;
}

BWORD TreeOp::rebuild(sp<SurfaceAccessibleI>& a_rspSurfaceAccessible)
{
	feLog("TreeOp::rebuild\n");

	setName("TreeOp (skeleton)");

	sp<SurfaceAccessorI> spInputVertices;
	if(!access(spInputVertices,a_rspSurfaceAccessible,e_primitive,e_vertices))
	{
		return FALSE;
	}

	m_forestRV.plants()->clear();

	Record plant=m_plantRV.createRecord();
	m_plantRV.plantName()="<skeleton>";
	m_plantRV.location()=convertY(catalog<SpatialVector>("TreeLocation"));

	sp<PlantModelI> spPlantModelI=registry()->create("PlantModelI.Tree");
	m_plantRV.plantModel()=spPlantModelI;

	const U32 primitiveCount=spInputVertices->count();
	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		const U32 subCount=spInputVertices->subCount(primitiveIndex);
		if(!subCount)
		{
			continue;
		}

		U32 fromIndex=spInputVertices->integer(primitiveIndex,0);

		SpatialVector fromPosition=convertY(
				spInputVertices->spatialVector(primitiveIndex,0));

		for(U32 subIndex=1;subIndex<subCount;subIndex++)
		{
			const U32 level=(primitiveIndex!=0);
			const U32 toIndex=spInputVertices->integer(primitiveIndex,subIndex);

			SpatialVector toPosition=convertY(
					spInputVertices->spatialVector(primitiveIndex,subIndex));

			Real radiusFade=magnitude(toPosition)/30.0;
			if(radiusFade>1.0)
			{
				radiusFade>1.0;
			}
			const Real radius=0.10-0.09*radiusFade;

			spPlantModelI->addBranch(fromIndex,toIndex,level,radius,
					toPosition-fromPosition);

			fromIndex=toIndex;
			fromPosition=toPosition;
		}
	}

	spPlantModelI->prepare();

	m_forestRV.plants()->add(plant);

	m_plantRV.bind(NULL);
	return TRUE;
}

void TreeOp::replace(sp<SurfaceAccessibleI>& a_rspSurfaceAccessible)
{
	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,a_rspSurfaceAccessible,e_point,e_position))
	{
		return;
	}

	sp<RecordGroup> spRG=m_forestRV.plants();
	RecordGroup::iterator it=spRG->begin();
	if(it==spRG->end())
	{
		return;
	}
	sp<RecordArray> spRA=*it;
	if(!spRA->length())
	{
		return;
	}
	Record plant=spRA->getRecord(0);
	if(!plant.isValid())
	{
		return;
	}
	m_plantRV.bind(plant);

	sp<PlantModelI> spPlantModelI=m_plantRV.plantModel();

	const U32 numSticks=spPlantModelI->numSticks();
	const U32 pointCount=spOutputPoint->count();
	if(pointCount)
	{
		SpatialVector root=unconvertY(m_plantRV.location());
		spOutputPoint->set(0,root);
	}
	for(U32 pointIndex=1;pointIndex<pointCount;pointIndex++)
	{
		SpatialVector tip;
		set(tip);

		if(pointIndex<=numSticks)
		{
			const PlantModelI::Stick* pStick=spPlantModelI->stick(pointIndex-1);
			const SpatialVector base=pStick->base();
			const SpatialVector span=pStick->span();
			tip=unconvertY(base+span);
		}
		spOutputPoint->set(pointIndex,tip);
	}

	m_plantRV.bind(NULL);
}

void TreeOp::regenerate(void)
{
	feLog("TreeOp::regenerate %s\n",m_treeName.c_str());

	setName(m_treeName);

	m_forestRV.plants()->clear();

	String longname;
	longname.sPrintf("../media/vegetation/%s.rg",m_treeName.c_str());
	sp<RecordGroup>	spRG=RecordView::loadRecordGroup(m_spScope,longname);
	RecordGroup::iterator it=spRG->begin();
	sp<RecordArray>& rspRA= *it;
	if(m_plantRV.plantName.check(rspRA))
	{
		m_forestRV.plants()->add(spRG);
	}
	else
	{
		Record plant=m_plantRV.createRecord();
		m_plantRV.plantName()=m_treeName;
		m_plantRV.location()=convertY(catalog<SpatialVector>("TreeLocation"));
		m_forestRV.plants()->add(plant);
	}

	m_plantRV.bind(NULL);
}

void TreeOp::handle(Record& a_rSignal)
{
	const Real frame=currentFrame(a_rSignal);
	const Real time=currentTime(a_rSignal);

//	feLog("TreeOp::handle frame %.6G time %.6G\n",frame,time);

	sp<SurfaceAccessibleI> spInputAccessible;
	access(spInputAccessible,"Input Surface",e_quiet);
	if(spInputAccessible.isValid())
	{
		m_treeName="";
		if(time<1e-6)
		{
			if(!rebuild(spInputAccessible))
			{
				return;
			}
		}
	}
	else
	{
		const String treeName=catalog<String>("TreeName");
		if(m_treeName!=treeName)
		{
			feLog("TreeOp::handle treeName changed \"%s\" -> \"%s\"\n",
					m_treeName.c_str(),treeName.c_str());
			m_treeName=treeName;
			regenerate();
		}
	}

	sp<DrawI> spDrawI;
	if(!accessDraw(spDrawI,a_rSignal)) return;

	m_forestRV.upAxis()=catalog<bool>("YUp")? 1: 2;
	m_forestRV.drawI()=spInputAccessible.isValid()? sp<DrawI>(NULL): spDrawI;
	m_forestRV.deltaT()=(1.0f/24.0f);
	m_forestRV.windDirection()=catalog<SpatialVector>("WindDirection");

	const SpatialVector center=catalog<SpatialVector>("ColliderCenter");
	const Real radius=catalog<Real>("ColliderRadius");
	Vector4 effector=convertY(center);
	effector[3]=radius;
	m_forestRV.effector()=effector;

	sp<RecordGroup>& rspPlants=m_forestRV.plants();
	FEASSERT(rspPlants.isValid());

	//* update location, but only if it is an individual tree
	for(RecordGroup::iterator it=rspPlants->begin();it!=rspPlants->end();it++)
	{
		sp<RecordArray> spRA= *it;

		if(spRA->length()!=1)
		{
			continue;
		}

		if(!m_plantRV.plantModel.check(spRA))
		{
			continue;
		}

		m_plantRV.bind(spRA);
		while(m_plantRV.step())
		{
			m_plantRV.location()=
					convertY(catalog<SpatialVector>("TreeLocation"));
		}
	}

//	m_plantRV.bind(NULL);

	Record forest=m_forestRV.record();
	m_spSignalerI->signal(forest);

	if(spInputAccessible.isValid())
	{
		sp<SurfaceAccessibleI> spOutputAccessible;
		accessOutput(spOutputAccessible,a_rSignal);

		replace(spOutputAccessible);
	}

//	feLog("TreeOp::handle done\n");
}
